@@ -41,7 +41,8 @@ typedef enum {
 	RSVG_DEF_RADGRAD,
 	RSVG_DEF_PATTERN,
 	RSVG_DEF_PATH,
-	RSVG_DEF_FILTER
+	RSVG_DEF_FILTER,
+	RSVG_DEF_MASK
 } RsvgDefType;
 
 struct _RsvgDefVal {
