@@ -36,6 +36,7 @@ typedef struct _RsvgGradientStop RsvgGradientStop;
 typedef struct _RsvgGradientStops RsvgGradientStops;
 typedef struct _RsvgLinearGradient RsvgLinearGradient;
 typedef struct _RsvgRadialGradient RsvgRadialGradient;
+typedef struct _RsvgPattern RsvgPattern;
 
 typedef struct _RsvgPaintServer RsvgPaintServer;
 
@@ -47,23 +48,14 @@ struct _RsvgPSCtx {
 	int x1;
 	int y1;
 
+	guint32 color;
 	double affine[6];
-	/* todo: we need to take in some context information, including:
-	   
-	1. The global affine transformation.
-	
-	2. User coordinates at time of reference (to implement
-	gradientUnits = "userSpaceOnUse").
-	
-	3. Object bounding box (to implement gradientUnits =
-	"objectBoundingBox").
-	
-	Maybe signal for lazy evaluation of object bbox.
-	*/
+	RsvgHandle *ctx;
 };
 
 struct _RsvgGradientStop {
 	double offset;
+	gboolean is_current_color;
 	guint32 rgba;
 };
 
@@ -80,6 +72,8 @@ struct _RsvgLinearGradient {
 	ArtGradientSpread spread;
 	double x1, y1;
 	double x2, y2;
+	guint32 current_color;
+	gboolean has_current_color;
 };
 
 struct _RsvgRadialGradient {
@@ -91,6 +85,18 @@ struct _RsvgRadialGradient {
 	double cx, cy;
 	double r;
 	double fx, fy;
+	guint32 current_color;
+	gboolean has_current_color;
+};
+
+struct _RsvgPattern {
+	RsvgDefVal super;
+	gboolean obj_cbbox;
+	gboolean obj_bbox;
+	double affine[6]; /* user space to actual at time of gradient def */
+	double x, y, width, height;
+	RsvgDefVal * g;
+	RsvgDefVal * gfallback;
 };
 
 /* Create a new paint server based on a specification string. */
@@ -114,6 +120,9 @@ rsvg_clone_radial_gradient (const RsvgRadialGradient *grad, gboolean * shallow_c
 RsvgLinearGradient *
 rsvg_clone_linear_gradient (const RsvgLinearGradient *grad, gboolean * shallow_cloned);
 
+RsvgPattern *
+rsvg_clone_pattern (const RsvgPattern *pattern);
+
 G_END_DECLS
 
 #endif
