@@ -66,6 +66,7 @@ rsvg_state_init (RsvgState *state)
 	art_affine_identity (state->personal_affine);
 	state->mask = NULL;
 	state->opacity = 0xff;
+	state->adobe_blend = 0;
 	state->fill = rsvg_paint_server_parse (NULL, NULL, "#000", 0);
 	state->fill_opacity = 0xff;
 	state->stroke_opacity = 0xff;
@@ -89,6 +90,7 @@ rsvg_state_init (RsvgState *state)
 	state->visible      = TRUE;
 	state->filter       = NULL;
 
+	state->has_current_color = FALSE;
 	state->has_fill_server = FALSE;
 	state->has_fill_opacity = FALSE;
 	state->has_fill_rule = FALSE;
@@ -139,6 +141,8 @@ rsvg_state_reinherit (RsvgState *dst, const RsvgState *src)
 {
 	gint i;
 	
+	if (!dst->has_current_color)
+		dst->current_color = src->current_color;
 	if (!dst->has_fill_server)
 		{
 			rsvg_paint_server_ref (src->fill);
@@ -211,6 +215,8 @@ rsvg_state_dominate (RsvgState *dst, const RsvgState *src)
 {
 	gint i;
 	
+	if (!dst->has_current_color || src->has_current_color)
+		dst->current_color = src->current_color;
 	if (!dst->has_fill_server || src->has_fill_server)
 		{
 			rsvg_paint_server_ref (src->fill);
@@ -238,7 +244,7 @@ rsvg_state_dominate (RsvgState *dst, const RsvgState *src)
 	if (!dst->has_join || src->has_join)
 		dst->join = src->join;
 	if (!dst->has_stop_color || src->has_stop_color)
-		dst->stop_color = src->stop_color;
+		dst->stop_color = src->stop_color;				
 	if (!dst->has_stop_opacity || src->has_stop_opacity)
 		dst->stop_opacity = src->stop_opacity;
 	if (!dst->has_visible || src->has_visible)
@@ -315,7 +321,7 @@ rsvg_parse_style_arg (RsvgHandle *ctx, RsvgState *state, const char *str)
 	if (rsvg_css_param_match (str, "color"))
 		{
 			RsvgState * parent_state = rsvg_state_parent (ctx);
-
+			
 			if (parent_state)
 				state->current_color = rsvg_css_parse_color (str + arg_off, parent_state->current_color);
 			else
@@ -329,6 +335,35 @@ rsvg_parse_style_arg (RsvgHandle *ctx, RsvgState *state, const char *str)
 		}
 	else if (rsvg_css_param_match (str, "filter"))
 		state->filter = rsvg_filter_parse(ctx->defs, str + arg_off);
+	else if (rsvg_css_param_match (str, "a:adobe-blending-mode"))
+		{
+			if (!strcmp (str + arg_off, "normal"))
+				state->adobe_blend = 0;
+			else if (!strcmp (str + arg_off, "multiply"))
+				state->adobe_blend = 1;
+			else if (!strcmp (str + arg_off, "screen"))
+				state->adobe_blend = 2;
+			else if (!strcmp (str + arg_off, "darken"))
+				state->adobe_blend = 3;
+			else if (!strcmp (str + arg_off, "lighten"))
+				state->adobe_blend = 4;
+			else if (!strcmp (str + arg_off, "softlight"))
+				state->adobe_blend = 5;
+			else if (!strcmp (str + arg_off, "hardlight"))
+				state->adobe_blend = 6;
+			else if (!strcmp (str + arg_off, "colordodge"))
+				state->adobe_blend = 7;
+			else if (!strcmp (str + arg_off, "colorburn"))
+				state->adobe_blend = 8;
+			else if (!strcmp (str + arg_off, "overlay"))
+				state->adobe_blend = 9;
+			else if (!strcmp (str + arg_off, "exclusion"))
+				state->adobe_blend = 10;
+			else if (!strcmp (str + arg_off, "difference"))
+				state->adobe_blend = 11;
+			else
+				state->adobe_blend = 0;
+		}
 	else if (rsvg_css_param_match (str, "mask"))
 		state->mask = rsvg_mask_parse(ctx->defs, str + arg_off);
 	else if (rsvg_css_param_match (str, "enable-background"))
@@ -540,14 +575,19 @@ rsvg_parse_style_arg (RsvgHandle *ctx, RsvgState *state, const char *str)
 	else if (rsvg_css_param_match (str, "stop-color"))
 		{
 			guint32 current_color = rsvg_state_current_color (state, parent_state);
-
-			state->has_stop_color = TRUE;
-			state->stop_color = rsvg_css_parse_color (str + arg_off, current_color);
+			if (strcmp(str + arg_off, "inherit"))
+				{
+					state->has_stop_color = TRUE;
+					state->stop_color = rsvg_css_parse_color (str + arg_off, current_color);
+				}
 		}
 	else if (rsvg_css_param_match (str, "stop-opacity"))
 		{
-			state->has_stop_opacity = TRUE;
-			state->stop_opacity = rsvg_css_parse_opacity (str + arg_off);
+			if (strcmp(str + arg_off, "inherit"))
+				{
+					state->has_stop_opacity = TRUE;
+					state->stop_opacity = rsvg_css_parse_opacity (str + arg_off);
+				}
 		}
 	else if (rsvg_css_param_match (str, "stroke-miterlimit"))
 		{
@@ -610,7 +650,7 @@ rsvg_parse_style_arg (RsvgHandle *ctx, RsvgState *state, const char *str)
 void rsvg_parse_style_pair (RsvgHandle *ctx, RsvgState *state, 
 							const char *key, const char *val)
 {
-	gchar * str = g_strdup_printf ("%s:%s", key, val);
+	gchar * str = g_strdup_printf ("%s:%s", key, val);;
 	rsvg_parse_style_arg (ctx, state, str);
 	g_free (str);
 }
@@ -629,6 +669,7 @@ void
 rsvg_parse_style_pairs (RsvgHandle *ctx, RsvgState *state, 
 						RsvgPropertyBag *atts)
 {
+			rsvg_lookup_parse_style_pair (ctx, state, "a:adobe-blending-mode", atts);
 			rsvg_lookup_parse_style_pair (ctx, state, "color", atts);
 			rsvg_lookup_parse_style_pair (ctx, state, "display", atts);
 			rsvg_lookup_parse_style_pair (ctx, state, "enable-background", atts);
@@ -759,19 +800,23 @@ ccss_end_selector (CRDocHandler *a_handler,
 }
 
 static void
-ccss_property (CRDocHandler *a_handler, GString *a_name, 
+ccss_property (CRDocHandler *a_handler, CRString *a_name, 
 			   CRTerm *a_expr, gboolean a_important)
 {
 	CSSUserData * user_data;
-	char * expr = NULL;
+	char * expr = NULL, *name = NULL;
+	size_t len = 0 ;
 
 	g_return_if_fail (a_handler);
 
 	user_data = (CSSUserData *)a_handler->app_data;
 
-	if (a_name && a_name->str && a_expr && user_data->def)
+	if (a_name && a_expr && user_data->def)
 		{
-			g_string_append_len (user_data->def, a_name->str, a_name->len);
+			name = (char*) cr_string_peek_raw_str (a_name) ;
+			len = cr_string_peek_raw_str_len (a_name) ;
+
+			g_string_append_len (user_data->def, name, len);
 			g_string_append (user_data->def, ": ");
 			expr = cr_term_to_string (a_expr);
 			g_string_append_len (user_data->def, expr, strlen (expr));
@@ -840,7 +885,6 @@ rsvg_real_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
         {
 			g_warning (_("Error setting CSS SAC handler"));
 			cr_parser_destroy (parser);
-			cr_doc_handler_destroy (css_handler);
 			return;
         }        
 	
@@ -848,7 +892,6 @@ rsvg_real_parse_cssbuffer (RsvgHandle *ctx, const char * buff, size_t buflen)
 	status = cr_parser_parse (parser);
 	
 	cr_parser_destroy (parser);
-	cr_doc_handler_destroy (css_handler);
 }
 
 #else /* !HAVE_LIBCROCO */
@@ -942,6 +985,9 @@ rsvg_parse_transform (double dst[6], const char *src)
 			while (g_ascii_isspace (src[idx]))
 				idx++;
 			
+			if (src[idx] == '\0')
+				break;
+
 			/* parse keyword */
 			for (key_len = 0; key_len < sizeof (keyword); key_len++)
 				{
@@ -1219,7 +1265,7 @@ rsvg_push_discrete_layer (RsvgHandle *ctx)
 	pixbuf = ctx->pixbuf;
 	
 	if (state->filter == NULL && state->opacity == 0xFF && 
-		!state->backgroundnew && state->mask == NULL)
+		!state->backgroundnew && state->mask == NULL && !state->adobe_blend)
 		return;
 
 	state->save_pixbuf = pixbuf;
@@ -1373,11 +1419,12 @@ rsvg_composite_layer(RsvgHandle *ctx, RsvgState *state, GdkPixbuf *tos, GdkPixbu
 	GdkPixbuf *intermediate;
 	GdkPixbuf *in, *out, *insidebg;
 	int operationsleft;
+	gint adobe_blend = state->adobe_blend;
 
 	intermediate = NULL;
 
 	if (state->filter == NULL && state->opacity == 0xFF && 
-		!state->backgroundnew && state->mask == NULL)
+		!state->backgroundnew && state->mask == NULL && !state->adobe_blend)
 		return;
 
 	operationsleft = 0;
@@ -1388,6 +1435,8 @@ rsvg_composite_layer(RsvgHandle *ctx, RsvgState *state, GdkPixbuf *tos, GdkPixbu
 		operationsleft++;
 	if (mask != NULL)
 		operationsleft++;
+	if (adobe_blend)
+		operationsleft++;		
 
 	if (operationsleft > 1)
 		intermediate = gdk_pixbuf_new (GDK_COLORSPACE_RGB, 1, 8, 
@@ -1401,12 +1450,17 @@ rsvg_composite_layer(RsvgHandle *ctx, RsvgState *state, GdkPixbuf *tos, GdkPixbu
 			rsvg_use_opacity (ctx, 0xFF, tos, nos);			
 		}
 
+	if (filter != NULL || adobe_blend)
+		{
+			insidebg = rsvg_compile_bg(ctx, state);
+		}
+	else
+		insidebg = NULL;
+
 	if (filter != NULL)
 		{
 			out = get_next_out(&operationsleft, in, tos, nos, intermediate);
-			insidebg = rsvg_compile_bg(ctx, state);
 			rsvg_filter_render (filter, in, out, insidebg, ctx);
-			g_object_unref (insidebg);
 			in = out;
 		}
 	if (opacity != 0xFF)
@@ -1421,6 +1475,17 @@ rsvg_composite_layer(RsvgHandle *ctx, RsvgState *state, GdkPixbuf *tos, GdkPixbu
 			rsvg_mask_render ((RsvgMask *)mask, in, out, ctx);
 			in = out;
 		}
+	if (adobe_blend)
+		{
+			out = get_next_out(&operationsleft, in, tos, nos, intermediate);
+			rsvg_filter_adobe_blend (adobe_blend, in, insidebg, out);
+			in = out;
+		}
+
+	if (filter != NULL || adobe_blend)
+		{
+			g_object_unref (insidebg);
+		}
 
 	if (intermediate != NULL)
 		g_object_unref (intermediate);
@@ -1444,7 +1509,7 @@ rsvg_pop_discrete_layer(RsvgHandle *ctx)
 	state = rsvg_state_current(ctx);
 
 	if (state->filter == NULL && state->opacity == 0xFF && 
-		!state->backgroundnew && state->mask == NULL)
+		!state->backgroundnew && state->mask == NULL && !state->adobe_blend)
 		return;
 
 	tos = ctx->pixbuf;
@@ -1457,6 +1522,12 @@ rsvg_pop_discrete_layer(RsvgHandle *ctx)
 	ctx->pixbuf = nos;
 }
 
+gboolean
+rsvg_needs_discrete_layer(RsvgState *state)
+{
+	return state->filter || state->mask || state->adobe_blend || state->backgroundnew;
+}
+
 RsvgState * 
 rsvg_state_current (RsvgHandle *ctx)
 {
