@@ -278,12 +278,13 @@ rsvg_gradient_stop_handler_start (RsvgSaxHandler *self, const xmlChar *name,
 	gboolean got_offset = FALSE;
 	RsvgState state;
 	int n_stop;
+	gboolean is_current_color = FALSE;
 	const char *value;
 	
 	if (strcmp ((char *)name, "stop"))
 		return;
 	
-	rsvg_state_init (&state);
+	rsvg_state_clone(&state ,rsvg_state_current(z->ctx));
 	
 	if (rsvg_property_bag_size (atts))
 		{
@@ -302,10 +303,13 @@ rsvg_gradient_stop_handler_start (RsvgSaxHandler *self, const xmlChar *name,
 			if ((value = rsvg_property_bag_lookup (atts, "style")))
 				rsvg_parse_style (z->ctx, &state, value);
 			
+			if ((value = rsvg_property_bag_lookup (atts, "stop-color")))
+				if (!strcmp(value, "currentColor"))
+					is_current_color = TRUE;
 			rsvg_parse_style_pairs (z->ctx, &state, atts);
 		}
 	
-	rsvg_state_finalize (&state);
+	rsvg_state_finalize(&state);
 	
 	if (!got_offset)
 		{
@@ -320,6 +324,7 @@ rsvg_gradient_stop_handler_start (RsvgSaxHandler *self, const xmlChar *name,
 		/* double the allocation if size is a power of two */
 		stops->stop = g_renew (RsvgGradientStop, stops->stop, n_stop << 1);
 	stops->stop[n_stop].offset = offset;
+	stops->stop[n_stop].is_current_color = is_current_color;
 	stops->stop[n_stop].rgba = (state.stop_color << 8) | state.stop_opacity;
 }
 
@@ -402,6 +407,8 @@ rsvg_start_linear_gradient (RsvgHandle *ctx, RsvgPropertyBag *atts)
 	gboolean obj_bbox = TRUE;
 	gboolean got_x1, got_x2, got_y1, got_y2, got_spread, got_transform, got_bbox, cloned, shallow_cloned;
 	double affine[6];
+	guint32 color = 0;
+	gboolean got_color = FALSE;
 	int i;
 
 	got_x1 = got_x2 = got_y1 = got_y2 = got_spread = got_transform = got_bbox = cloned = shallow_cloned = FALSE;
@@ -445,11 +452,18 @@ rsvg_start_linear_gradient (RsvgHandle *ctx, RsvgPropertyBag *atts)
 				xlink_href = value;
 			if ((value = rsvg_property_bag_lookup (atts, "gradientTransform")))
 				got_transform = rsvg_parse_transform (affine, value);
+			if ((value = rsvg_property_bag_lookup (atts, "color")))
+				{
+					got_color = TRUE;
+					color = rsvg_css_parse_color (value, 0);
+				}
 			if ((value = rsvg_property_bag_lookup (atts, "gradientUnits"))) {
 				if (!strcmp (value, "userSpaceOnUse"))
 					obj_bbox = FALSE;
 				got_bbox = TRUE;
+
 			}
+			rsvg_parse_style_pairs (ctx, state, atts);
 		}
 	
 	/* set up 100% as the default if not gotten */
@@ -487,12 +501,28 @@ rsvg_start_linear_gradient (RsvgHandle *ctx, RsvgPropertyBag *atts)
 	else
 		art_affine_identity(grad->affine);
 
+	if (got_color)
+		{
+			grad->current_color = color;
+			grad->has_current_color = TRUE;
+		}
+
 	/* gradient inherits parent/cloned information unless it's explicity gotten */
 	grad->obj_bbox = (cloned && !got_bbox) ? grad->obj_bbox : obj_bbox;
-	grad->x1 = (cloned && !got_x1) ? grad->x1 : x1;
-	grad->y1 = (cloned && !got_y1) ? grad->y1 : y1;
-	grad->x2 = (cloned && !got_x2) ? grad->x2 : x2;
-	grad->y2 = (cloned && !got_y2) ? grad->y1 : y2;
+	if (!shallow_cloned)
+		{
+			grad->x1 = (cloned && !got_x1) ? grad->x1 : x1;
+			grad->y1 = (cloned && !got_y1) ? grad->y1 : y1;
+			grad->x2 = (cloned && !got_x2) ? grad->x2 : x2;
+			grad->y2 = (cloned && !got_y2) ? grad->y2 : y2;
+		}
+	else
+		{
+			grad->x1 = x1;
+			grad->y1 = y1;
+			grad->x2 = x2;
+			grad->y2 = y2;
+		}
 	grad->spread = (cloned && !got_spread) ? grad->spread : spread;
 }
 
@@ -518,6 +548,8 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, RsvgPropertyBag *atts, const char *
 	ArtGradientSpread spread = ART_GRADIENT_PAD;
 	gboolean obj_bbox = TRUE;
 	gboolean got_cx, got_cy, got_r, got_fx, got_fy, got_spread, got_transform, got_bbox, cloned, shallow_cloned;
+	guint32 color = 0;
+	gboolean got_color = FALSE;
 	double affine[6];
 	int i;
 
@@ -553,6 +585,11 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, RsvgPropertyBag *atts, const char *
 			if ((value = rsvg_property_bag_lookup (atts, "gradientTransform"))) {
 				got_transform = rsvg_parse_transform (affine, value);
 			}
+			if ((value = rsvg_property_bag_lookup (atts, "color")))
+				{
+					got_color = TRUE;
+					color = rsvg_css_parse_color (value, 0);
+				}
 			if ((value = rsvg_property_bag_lookup (atts, "spreadMethod")))
 				{
 					if (!strcmp (value, "pad")) {
@@ -573,6 +610,7 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, RsvgPropertyBag *atts, const char *
 					obj_bbox = FALSE;
 				got_bbox = TRUE;
 			}
+			rsvg_parse_style_pairs (ctx, state, atts);
 		}
 	
 	if (xlink_href != NULL)
@@ -627,16 +665,146 @@ rsvg_start_radial_gradient (RsvgHandle *ctx, RsvgPropertyBag *atts, const char *
 	else
 		art_affine_identity(grad->affine);
 	
+	if (got_color)
+		{
+			grad->current_color = color;
+			grad->has_current_color = TRUE;
+		}
+
 	/* gradient inherits parent/cloned information unless it's explicity gotten */
 	grad->obj_bbox = (cloned && !got_bbox) ? grad->obj_bbox : obj_bbox;
-	grad->cx = (cloned && !got_cx) ? grad->cx : cx;
-	grad->cy = (cloned && !got_cy) ? grad->cy : cy;
-	grad->r =  (cloned && !got_r)  ? grad->r  : r;
-	grad->fx = (cloned && !got_fx) ? grad->fx : fx;
-	grad->fy = (cloned && !got_fy) ? grad->fy : fy;
+	if (!shallow_cloned)
+		{
+			grad->cx = (cloned && !got_cx) ? grad->cx : cx;
+			grad->cy = (cloned && !got_cy) ? grad->cy : cy;
+			grad->r =  (cloned && !got_r)  ? grad->r  : r;
+			grad->fx = (cloned && !got_fx) ? grad->fx : fx;
+			grad->fy = (cloned && !got_fy) ? grad->fy : fy;
+		}
+	else
+		{
+			grad->cx = cx;
+			grad->cy = cy;
+			grad->r = r;
+			grad->fx = fx;
+			grad->fy = fy;
+		}
 	grad->spread = (cloned && !got_spread) ? grad->spread : spread;
 }
 
+void
+rsvg_pattern_free (RsvgDefVal *self)
+{
+	RsvgPattern *z = (RsvgPattern *)self;
+	
+	g_free (z);
+}
+
+static void
+rsvg_start_pattern (RsvgHandle *ctx, RsvgPropertyBag *atts)
+{
+	RsvgState *state = rsvg_state_current (ctx);
+	RsvgPattern *pattern = NULL;
+	const char *id = NULL, *value;
+	double x = 0., y = 0., width = 0., height = 0.;
+	const char * xlink_href = NULL;
+	gboolean obj_bbox = TRUE;
+	gboolean obj_cbbox = FALSE;
+	gboolean got_x, got_y, got_width, got_height, got_transform, got_bbox, got_cbbox, cloned;
+	double affine[6];
+	int i;
+
+	got_x = got_y = got_width = got_height = got_transform = got_bbox = got_cbbox = cloned = FALSE;
+		
+	if (rsvg_property_bag_size (atts))
+		{
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+			if ((value = rsvg_property_bag_lookup (atts, "x"))) {
+				x = rsvg_css_parse_normalized_length (value, ctx->dpi, 1, state->font_size);
+				got_x = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "y"))) {
+				y = rsvg_css_parse_normalized_length (value, ctx->dpi, 1, state->font_size);
+				got_y = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "width"))) {
+				width = rsvg_css_parse_normalized_length (value, ctx->dpi, 1, state->font_size);
+				got_width = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "height"))) {
+				height = rsvg_css_parse_normalized_length (value, ctx->dpi, 1, state->font_size);
+				got_height = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "xlink:href")))
+				xlink_href = value;
+			if ((value = rsvg_property_bag_lookup (atts, "patternTransform")))
+				got_transform = rsvg_parse_transform (affine, value);
+			if ((value = rsvg_property_bag_lookup (atts, "patternUnits"))) {
+				if (!strcmp (value, "userSpaceOnUse"))
+					obj_bbox = FALSE;
+				else
+					obj_bbox = TRUE;
+				got_bbox = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "patternContentUnits"))) {
+				if (!strcmp (value, "userSpaceOnUse"))
+					obj_cbbox = FALSE;
+				else
+					obj_cbbox = TRUE;					
+				got_cbbox = TRUE;
+			}
+		}
+	
+	/* set up 100% as the default if not gotten */
+	
+	if (!got_width) {
+		if (obj_bbox)
+			width = 1.0;
+		else
+			width = rsvg_css_parse_normalized_length ("100%", ctx->dpi, (gdouble)ctx->width, state->font_size);
+	}
+
+	if (xlink_href != NULL)
+		{
+			RsvgPattern * parent = (RsvgPattern*)rsvg_defs_lookup (ctx->defs, xlink_href+1);
+			if (parent != NULL)
+				{
+					cloned = TRUE;
+					pattern = rsvg_clone_pattern (parent);
+				}
+		}
+	
+	if (!cloned)
+		{
+			pattern = g_new (RsvgPattern, 1);
+			pattern->super.type = RSVG_DEF_PATTERN;
+			pattern->super.free = rsvg_pattern_free;
+			pattern->gfallback = NULL;
+		}
+	
+	rsvg_defs_set (ctx->defs, id, &pattern->super);
+	
+	if (got_transform)
+		for (i = 0; i < 6; i++)
+			pattern->affine[i] = affine[i];
+	else
+		art_affine_identity(pattern->affine);
+
+	/* gradient inherits parent/cloned information unless it's explicity gotten */
+	pattern->obj_bbox = (cloned && !got_bbox) ? pattern->obj_bbox : obj_bbox;
+	pattern->obj_cbbox = (cloned && !got_cbbox) ? pattern->obj_cbbox : obj_cbbox;
+	pattern->x = (cloned && !got_x) ? pattern->x : x;
+	pattern->y = (cloned && !got_y) ? pattern->y : y;
+	pattern->width = (cloned && !got_width) ? pattern->width : width;
+	pattern->height = (cloned && !got_height) ? pattern->height : height;
+
+	ctx->in_defs++;		
+
+	pattern->g = &(rsvg_push_def_group (ctx, "")->super);
+}
+
+
 /* end gradients */
 
 static void
@@ -698,18 +866,6 @@ rsvg_start_style (RsvgHandle *ctx, RsvgPropertyBag *atts)
 	ctx->handler = &handler->super;
 }
 
-/* start defs */
-
-static void
-rsvg_defs_handler_free (RsvgSaxHandler *self)
-{
-	g_free (self);
-}
-
-static void
-rsvg_defs_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
-{
-}
 
 static void
 rsvg_filter_handler_start (RsvgHandle *ctx, const xmlChar *name,
@@ -767,109 +923,6 @@ rsvg_filter_handler_start (RsvgHandle *ctx, const xmlChar *name,
 		rsvg_start_filter_primitive_component_transfer_function(ctx, atts, 'a');
 }
 
-static void
-rsvg_defs_handler_start (RsvgSaxHandler *self, const xmlChar *name,
-						 RsvgPropertyBag *atts)
-{
-	RsvgSaxHandlerDefs *z = (RsvgSaxHandlerDefs *)self;
-	RsvgHandle *ctx = z->ctx;
-	
-	/* push the state stack */
-	if (ctx->n_state == ctx->n_state_max)
-		ctx->state = g_renew (RsvgState, ctx->state, ctx->n_state_max <<= 1);
-	if (ctx->n_state)
-		{
-			rsvg_state_inherit (&ctx->state[ctx->n_state],
-								&ctx->state[ctx->n_state - 1]);
-		}
-	else
-		rsvg_state_init (ctx->state);
-	ctx->n_state++;
-
-	/*
-	 * conicalGradient isn't in the SVG spec and I'm not sure exactly what it does. libart definitely
-	 * has no analogue. But it does seem similar enough to a radialGradient that i'd rather get the
-	 * onscreen representation of the colour wrong than not have any colour displayed whatsoever
-	 */
-
-	if (!strcmp ((char *)name, "defs"))
-		ctx->in_defs++;
-	else if (!strcmp ((char *)name, "linearGradient"))
-		rsvg_start_linear_gradient (ctx, atts);
-	else if (!strcmp ((char *)name, "radialGradient"))
-		rsvg_start_radial_gradient (ctx, atts, "radialGradient");
-	else if (!strcmp((char *)name, "conicalGradient"))
-		rsvg_start_radial_gradient (ctx, atts, "conicalGradient");
-	else if (!strcmp ((char *)name, "style"))
-		rsvg_start_style (ctx, atts);
-	else if (!strcmp ((char *)name, "g"))
-		rsvg_start_g (ctx, atts);
-	else if (!strcmp ((char *)name, "use"))
-		rsvg_start_use (ctx, atts);
-	else if (!strcmp ((char *)name, "path"))
-		rsvg_start_path (ctx, atts);
-	else if (!strcmp ((char *)name, "line"))
-		rsvg_start_line (ctx, atts);
-	else if (!strcmp ((char *)name, "rect"))
-		rsvg_start_rect (ctx, atts);
-	else if (!strcmp ((char *)name, "circle"))
-		rsvg_start_circle (ctx, atts);
-	else if (!strcmp ((char *)name, "ellipse"))
-		rsvg_start_ellipse (ctx, atts);
-	else if (!strcmp ((char *)name, "polygon"))
-		rsvg_start_polygon (ctx, atts);
-	else if (!strcmp ((char *)name, "polyline"))
-		rsvg_start_polyline (ctx, atts);
-	else if (!strcmp ((char *)name, "mask"))
-		rsvg_start_mask(ctx, atts);	
-	rsvg_filter_handler_start (ctx, name, atts);
-}
-
-static void
-rsvg_defs_handler_end (RsvgSaxHandler *self, const xmlChar *name)
-{
-	RsvgSaxHandlerDefs *z = (RsvgSaxHandlerDefs *)self;
-	RsvgHandle *ctx = z->ctx;
-	
-	if (!strcmp((char *)name, "defs"))
-		{
-			if (ctx->handler != NULL)
-				{
-					ctx->handler->free (ctx->handler);
-					ctx->handler = NULL;
-				}
-			ctx->in_defs--;
-		}
-
-	if (!strcmp ((char *)name, "g"))
-		rsvg_end_g (ctx);
-	if (!strcmp ((char *)name, "filter"))
-		rsvg_end_filter (ctx);
-	else if (!strcmp ((char *)name, "mask"))
-		rsvg_end_mask(ctx);
-	
-	/* pop the state stack */
-	ctx->n_state--;
-	rsvg_state_finalize (&ctx->state[ctx->n_state]);
-}
-
-static void
-rsvg_start_defs (RsvgHandle *ctx, RsvgPropertyBag *atts)
-{
-	RsvgSaxHandlerDefs *handler = g_new0 (RsvgSaxHandlerDefs, 1);
-	
-	handler->super.free = rsvg_defs_handler_free;
-	handler->super.characters = rsvg_defs_handler_characters;
-	handler->super.start_element = rsvg_defs_handler_start;
-	handler->super.end_element   = rsvg_defs_handler_end;
-	handler->ctx = ctx;
-
-	ctx->in_defs++;	
-	ctx->handler = &handler->super;
-}
-
-/* end defs */
-
 /* start desc */
 
 static void
@@ -896,7 +949,7 @@ rsvg_desc_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	string = g_strndup (ch, len);
 	if (!g_utf8_validate (string, -1, NULL))
 		{
-			utf8 = make_valid_utf8 (string);
+			utf8 = rsvg_make_valid_utf8 (string);
 			g_free (string);
 			string = utf8;
 		}
@@ -973,7 +1026,7 @@ rsvg_title_handler_characters (RsvgSaxHandler *self, const xmlChar *ch, int len)
 	string = g_strndup (ch, len);
 	if (!g_utf8_validate (string, -1, NULL))
 		{
-			utf8 = make_valid_utf8 (string);
+			utf8 = rsvg_make_valid_utf8 (string);
 			g_free (string);
 			string = utf8;
 		}
@@ -1023,9 +1076,9 @@ rsvg_start_title (RsvgHandle *ctx, RsvgPropertyBag *atts)
 }
 
 /* end title */
-
 static void
-rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
+rsvg_start_element (void *data, const xmlChar *name,
+					const xmlChar ** atts)
 {
 	RsvgHandle *ctx = (RsvgHandle *)data;
 
@@ -1057,8 +1110,13 @@ rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
 				rsvg_start_svg (ctx, bag);
 			else if (!strcmp ((char *)name, "g"))
 				rsvg_start_g (ctx, bag);
+			else if (!strcmp ((char *)name, "symbol"))
+				{
+					ctx->in_defs++;
+					rsvg_start_g (ctx, bag);
+				}
 			else if (!strcmp ((char *)name, "defs"))
-				rsvg_start_defs (ctx, bag);
+				ctx->in_defs++;	
 			else if (!strcmp ((char *)name, "path"))
 				rsvg_start_path (ctx, bag);
 			else if (!strcmp ((char *)name, "line"))
@@ -1098,6 +1156,8 @@ rsvg_start_element (void *data, const xmlChar *name, const xmlChar **atts)
 				rsvg_start_radial_gradient (ctx, bag, "radialGradient");
 			else if (!strcmp ((char *)name, "conicalGradient"))
 				rsvg_start_radial_gradient (ctx, bag, "conicalGradient");
+			else if (!strcmp ((char *)name, "pattern"))
+				rsvg_start_pattern (ctx, bag);
 
 			rsvg_filter_handler_start (ctx, name, bag);
     }
@@ -1126,7 +1186,12 @@ rsvg_end_element (void *data, const xmlChar *name)
 
 			if (!strcmp ((char *)name, "g"))
 				rsvg_end_g (ctx);
-			if (!strcmp ((char *)name, "filter"))
+			else if (!strcmp ((char *)name, "symbol"))
+				{
+					ctx->in_defs--;
+					rsvg_end_g (ctx);
+				}
+			else if (!strcmp ((char *)name, "filter"))
 				rsvg_end_filter (ctx);
 			else if (!strcmp ((char *)name, "defs")) {
 				ctx->in_defs--;	
@@ -1135,6 +1200,10 @@ rsvg_end_element (void *data, const xmlChar *name)
 				rsvg_end_mask(ctx);
 				ctx->in_defs--;				
 			}
+			else if (!strcmp ((char *)name, "pattern")) {
+				ctx->in_defs--;	
+				rsvg_pop_def_group(ctx);
+			}
 			/* pop the state stack */
 			ctx->n_state--;
 			rsvg_state_finalize (&ctx->state[ctx->n_state]);
@@ -1192,36 +1261,19 @@ rsvg_error_cb (void *data, const char *msg, ...)
 	va_end (args);
 }
 
-static xmlSAXHandler rsvgSAXHandlerStruct = {
-    NULL, /* internalSubset */
-    NULL, /* isStandalone */
-    NULL, /* hasInternalSubset */
-    NULL, /* hasExternalSubset */
-    NULL, /* resolveEntity */
-    rsvg_get_entity, /* getEntity */
-    rsvg_entity_decl, /* entityDecl */
-    NULL, /* notationDecl */
-    NULL, /* attributeDecl */
-    NULL, /* elementDecl */
-    NULL, /* unparsedEntityDecl */
-    NULL, /* setDocumentLocator */
-    NULL, /* startDocument */
-    NULL, /* endDocument */
-    rsvg_start_element, /* startElement */
-    rsvg_end_element, /* endElement */
-    NULL, /* reference */
-    rsvg_characters, /* characters */
-    NULL, /* ignorableWhitespace */
-    NULL, /* processingInstruction */
-    NULL, /* comment */
-    NULL, /* xmlParserWarning */
-    rsvg_error_cb, /* xmlParserError */
-    rsvg_error_cb, /* xmlParserFatalError */
-    NULL, /* getParameterEntity */
-    rsvg_characters, /* cdataCallback */
-    NULL /* externalSubset */
-	/* initialized */
-};
+static xmlSAXHandler rsvgSAXHandlerStruct;
+
+static void rsvg_SAX_handler_struct_init()
+{
+	rsvgSAXHandlerStruct.getEntity = rsvg_get_entity;
+    rsvgSAXHandlerStruct.entityDecl = rsvg_entity_decl;
+    rsvgSAXHandlerStruct.characters = rsvg_characters;
+    rsvgSAXHandlerStruct.error = rsvg_error_cb;
+	rsvgSAXHandlerStruct.error = rsvg_error_cb;
+	rsvgSAXHandlerStruct.cdataBlock = rsvg_characters;
+    rsvgSAXHandlerStruct.startElement = rsvg_start_element;
+    rsvgSAXHandlerStruct.endElement = rsvg_end_element;
+}
 
 /**
  * rsvg_error_quark
@@ -1398,6 +1450,7 @@ rsvg_handle_init (RsvgHandle * handle)
 	
 	handle->css_props = g_hash_table_new_full (g_str_hash, g_str_equal,
 											   g_free, g_free);
+	rsvg_SAX_handler_struct_init();
 	
 	handle->ctxt = NULL;
 	handle->current_defs_group = NULL;
